#
#   FP_prop.py
#
#
#   Created by David Limmer on 10/7/19
#
# Copyright 2021 David T Limmer
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
#
# 1. Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer. 
#
# 2. Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in
#   the documentation and/or other materials provided with the
#   distribution. 
#
# 3. Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission. 
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
#   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
#   FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
#   COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
#   INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
#   (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#   SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
#   OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
#   EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pylab import *

# Natural units are A and ps
# Set up domain
dx=0.2
xmin=-300.
xmax=30.
x=arange(xmin,xmax,dx)
# Timestep for stability
dt=0.45*dx**2.
# Total number of integration steps
steps=15000000

xm=len(x)-2

x1=int((12.0-xmin)/dx)
x2=int((16.0-xmin)/dx)

# Definition of interface
xrb=int((9.0-xmin)/dx)
xri=int((17.0-xmin)/dx)

# Definition of vapor x>xe
xe=int((19.-xmin)/dx)

# Parameterization of free energy function in kbT
c1= 1.28267/.5922
c2= 18.115
c3= 2.01285
c4= 2.41484
c5= 9.73522
c6= 13.9676
c7= 1.18626/.5922
c8= 0.607262/.5922#+5.5
c9=2.30687
c10= 5.30734

# Free energy
U=c1*tanh((x-c2)/c3)+c4-c7*exp(-(x-c6)**2./2./c10)+exp(-(x-c5)**2./2./c9)*c8
# First derivative
Up=-(-c1*1./cosh((x-c2)/c3)**2./c3+c7*exp(-(x-c6)**2./2./c10)*(c6-x)/c10+c8*exp(-(x-c5)**2./2./c9)*(-c5+x)/c9)
# Second derivative
Upp=-(2*c1*1./cosh((x-c2)/c3)**2.*tanh((x-c2)/c3)/c3**2.+c7*exp(-(x-c6)**2./2./c10)*(-c10+(c6-x)**2.)/c10**2.+c8*exp(-(x-c5)**2./2./c9)*(c9-(c5-x)**2.)/c9**2.)


#fraction of interfacial rate 1/delta
delt=5.

# Bulk Rate in units of per dt
kss=array([0.000005,0.00001,0.00002,0.00003,0.00004,0.00005,0.00006,0.00007,0.00008,0.00009])

for k in kss:
    # Initial condition
    P_0=exp(-(x-14.)**2./2.)
    P_0[:x1]=0
    P_0[x2:]=0
    P_0=P_0/sum(P_0)
    P_i=zeros(len(P_0))
    P_ip=zeros(len(P_0))

    # Interface rate
    ks=k/delt
    i=0
    # Save to file
    output=open('New-uptake_100k%f_100ks%f_del_%f.txt' %(k*100,ks*100,delt),'w')

    D=zeros(len(P_i))
    D[:xri]=0.16
    D[xri:]=0.16*3.

    for t in range(steps+1):
        # Fokker Plank update using finite difference

        P_i[2:xm-1]=P_0[2:xm-1]+D[2:xm-1]*dt*((P_0[1:xm-2]-2*P_0[2:xm-1]+P_0[3:xm])/dx**2+Upp[2:xm-1]*P_0[2:xm-1]+Up[2:xm-1]*(P_0[3:xm]-P_0[1:xm-2])/dx/2.)

        # Reflecting BCs
        P_i[2]=-P_ip[2]+2.*P_0[2]+2.*(P_0[3]-P_0[2])/dx**2.*dt*dt
        P_i[xm-1]=-P_ip[xm-1]+2.*P_0[xm-1]+2.*(P_0[xm-2]-P_0[xm-1])/dx**2.*dt*dt
        
        # Hydrolysis
        P_i[3:xrb]=P_i[3:xrb]-P_0[3:xrb]*dt*k
        
        P_i[xrb:xri]=P_i[xrb:xri]-P_0[xrb:xri]*dt*ks

        
        P_i[0]=P_0[0]+sum(P_0[3:xrb]*dt*k)+sum(P_0[xrb:xri]*dt*ks)
        
        # Evaporation
        P_i[-1]=P_0[-1]+sum(P_i[xe:xm])
        P_i[xe:xm]=0.
        
        # Ensure normalization
        P_i=P_i/sum(P_i)
        
        # Plot uptake as a function of time
        if(t%1000==0):
            # time/ps, hydrolysis concentration, evaporation concentration, reactive uptake gamma
            print t, P_i[0], P_i[xm+1], 1./(1+P_i[xm+1]/P_i[0])
            print >> output, t, P_i[0], P_i[xm+1], 1./(1+P_i[xm+1]/P_i[0])


        P_ip=P_0.copy()
        P_0=P_i.copy()


